from transformers import AutoTokenizer, AutoModel, AutoModelForSequenceClassification, AutoModelForTokenClassification
import torch
import numpy as np
from toolz import partition_all
from tqdm import tqdm
from scales_nlp.utils import convert_default_binary_outputs, load_config


def pipeline(pipeline_name, task=None, **kwargs):
    task_pipelines = {
        'classification': ClassificationPipeline,
        'multi-label-classification': MultiLabelClassificationPipeline,
        'ner': TokenClassificationPipeline,
        'token-classification': TokenClassificationPipeline,
        'sentence-encoding': SentenceEncodingPipeline,
    }

    pipelines = {
        'ontology': OntologySingleLabelPipeline,
        'docket-triplet-encoder': DocketTripletPipeline,
    }

    if task is not None:
        if task in task_pipelines:
            return task_pipelines[task](pipeline_name, **kwargs)
        else:
            raise Exception("'%s' is not a valid task name.  This should be one of 'classification', 'multi-label-classification'" % task)

    if pipeline_name in pipelines:
            return pipelines[pipeline_name](pipeline_name, **kwargs)
    else:
        raise Exception("'%s' is not a valid pipeline name.  This should be one of 'ontology'" % pipeline_name)


class BasePipeline(object):
    def __init__(self, model_name, max_length=512, device=None, use_auth_token=False, **kwargs):
        config = load_config()

        use_auth_token = use_auth_token if not self.require_auth_token else True
        if use_auth_token == True and config['HUGGING_FACE_TOKEN'] is not None:
            use_auth_token = config['HUGGING_FACE_TOKEN']

        self.device = device if device is not None else -1 if not torch.cuda.is_available() else torch.cuda.current_device()
        self.max_length = max_length
        self.use_auth_token = use_auth_token
        self.model_name = self.get_model_name(model_name, **kwargs)
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, use_auth_token=self.use_auth_token)
        self.model = self.load_model()
    
    def place_on_device(self, obj):
        if self.device >= 0:
            return obj.to(self.device)
        else:
            return obj.to('cpu')
    
    def get_model_name(self, model_name, **kwargs):
        return model_name

    def process_inputs(self, examples):
        return examples
    
    def process_predictions(self, predictions):
        return predictions
    
    def tokenize(self, texts):
        return self.tokenizer(texts, padding='max_length', max_length=self.max_length, truncation=True, return_tensors='pt')
    
    def generate_batches(self, examples, batch_size):
        return [list(batch) for batch in partition_all(batch_size, examples)]
        
    def load_model(self):
        raise NotImplementedError('load_model not implemented')
    
    def batch_predict(self, inputs, **kwargs):
        raise NotImplementedError('batch_predict not implemented')

    def __call__(self, examples, batch_size=4, verbose=True, **kwargs):
        examples = self.process_inputs(examples)
        batches = self.generate_batches(examples, batch_size)
        predictions = []
        for batch in tqdm(batches, disable=not verbose):
            inputs = self.place_on_device(self.tokenize(batch))
            predictions += self.batch_predict(inputs, **kwargs)
        return self.process_predictions(predictions)

    @property
    def require_auth_token(self):
        return False


# Task Pipelines
class ClassificationPipeline(BasePipeline):
    def load_model(self):
        model = AutoModelForSequenceClassification.from_pretrained(self.model_name, use_auth_token=self.use_auth_token)
        model = self.place_on_device(model)
        return model
    
    def batch_predict(self, inputs, return_scores=False, **kwargs):
        outputs = self.model(**inputs)
        logits = outputs.logits.detach().cpu()
        scores = torch.softmax(logits, dim=-1)
        predictions = []
        for i in range(len(scores)):
            prediction = {self.model.config.id2label[label_id]: score.item() for label_id, score in enumerate(scores[i])}
            if not return_scores:
                prediction = max(prediction.items(), key=lambda x: x[1])[0]
            predictions.append(prediction)
        return predictions


class MultiLabelClassificationPipeline(ClassificationPipeline):
    def batch_predict(self, inputs, return_scores=False, prediction_threshold=0.5, **kwargs):
        outputs = self.model(**inputs)
        logits = outputs.logits.detach().cpu()
        scores = torch.sigmoid(logits)
        predictions = []
        for i in range(len(scores)):
            prediction = {self.model.config.id2label[label_id]: score.item() for label_id, score in enumerate(scores[i])}
            if not return_scores:
                prediction = [label for label, score in prediction.items() if score > prediction_threshold]
            predictions.append(prediction)
        return predictions


class TokenClassificationPipeline(BasePipeline):
    def tokenize(self, texts):
        return self.tokenizer(texts, 
            padding='max_length', 
            max_length=self.max_length, 
            truncation=True, 
            return_offsets_mapping=True,
            return_tensors='pt'
        )

    def load_model(self):
        model = AutoModelForTokenClassification.from_pretrained(self.model_name, use_auth_token=self.use_auth_token)
        model = self.place_on_device(model)
        return model
    
    def batch_predict(self, inputs, **kwargs):
        offsets = inputs.pop('offset_mapping')
        outputs = self.model(**inputs)
        logits = outputs.logits.detach().cpu()
        scores = torch.softmax(logits, dim=-1)
        predictions = []
        for i in range(len(scores)):
            prediction = []
            example_offsets = offsets[i]
            input_ids = inputs['input_ids'][i]
            example_scores = scores[i][inputs['attention_mask'][i] == 1]
            example_scores, labels = torch.max(example_scores, dim=-1)
            labels = [self.model.config.id2label[x.item()] for x in labels]
            for token_idx in range(len(labels)):
                label = labels[token_idx].split('-')
                pos, entity = label[0], label[-1]
                if pos == 'B' or (pos == 'I' and (len(prediction) == 0 or prediction[-1]['entity'] != entity)):
                    prediction.append({
                        'entity': entity,
                        'start': example_offsets[token_idx][0].item(),
                        'end': example_offsets[token_idx][1].item(),
                        'score': [example_scores[token_idx].item()],
                        'text': [input_ids[token_idx]]
                    })
                if pos == 'I':
                    prediction[-1]['end'] = example_offsets[token_idx][1].item()
                    prediction[-1]['score'].append(example_scores[token_idx].item())
                    prediction[-1]['text'].append(input_ids[token_idx])
            for span in prediction:
                span['score'] = np.mean(span['score'])
                span['text'] = self.tokenizer.decode(span['text'])
            predictions.append(prediction)
        return predictions


class SentenceEncodingPipeline(BasePipeline):
    def load_model(self):
        model = AutoModel.from_pretrained(self.model_name, use_auth_token=self.use_auth_token)
        model = self.place_on_device(model)
        return model
    
    def batch_predict(self, inputs, **kwargs):
        outputs = self.model(**inputs)
        return [outputs[0].detach().cpu().numpy()[:,0,:]]
    
    def process_predictions(self, predictions):
        return np.concatenate(predictions, axis=0)


# Custom Pipelines
class OntologySingleLabelPipeline(ClassificationPipeline):
    def get_model_name(self, model_name, **kwargs):
        return 'scales-okn/ontology-' + kwargs['label_name'].replace(' ', '-')

    def process_predictions(self, predictions):
        return convert_default_binary_outputs(predictions)
    
    @property
    def require_auth_token(self):
        return True


class DocketTripletPipeline(SentenceEncodingPipeline):
    def get_model_name(self, model_name, **kwargs):
        return 'scales-okn/docket-triplet-vectors'
    
    @property
    def require_auth_token(self):
        return True


